import datetime
import os
import platform
import sys
import typing
from collections.abc import Mapping

import _pytest.config
import _pytest.config.argparsing
import _pytest.main
import _pytest.nodes
import _pytest.pathlib
import _pytest.reports
import _pytest.runner
import _pytest.terminal
import opentelemetry.trace
import pytest
import pytest_timeout
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator

from pytest_mergify import utils
from pytest_mergify.ci_insights import MergifyCIInsights


class PytestMergify:
    mergify_ci: MergifyCIInsights

    def pytest_configure(self, config: _pytest.config.Config) -> None:
        kwargs = {}
        api_url = config.getoption("--mergify-api-url")
        if api_url is not None:
            kwargs["api_url"] = api_url
        self.mergify_ci = MergifyCIInsights(**kwargs)

    def pytest_terminal_summary(
        self, terminalreporter: _pytest.terminal.TerminalReporter
    ) -> None:
        # No CI, nothing to do
        if not utils.is_in_ci():
            return

        terminalreporter.section("Mergify CI")

        if not self.mergify_ci.token:
            terminalreporter.write_line(
                "No token configured for Mergify; test results will not be uploaded",
                yellow=True,
            )
            return

        if not self.mergify_ci.repo_name:
            terminalreporter.write_line(
                "Unable to determine repository name; test results will not be uploaded",
                red=True,
            )
            return

        if self.mergify_ci.flaky_detector:
            terminalreporter.write_line(self.mergify_ci.flaky_detector.make_report())
        elif self.mergify_ci.flaky_detector_error_message:
            terminalreporter.write_line(
                f"""⚠️ Flaky detection couldn't be enabled because of an error.

Common issues:
  • Your 'MERGIFY_TOKEN' might not be set or could be invalid
  • There might be a network connectivity issue with the Mergify API

📚 Documentation: https://docs.mergify.com/ci-insights/test-frameworks/pytest/
🔍 Details: {self.mergify_ci.flaky_detector_error_message}""",
                yellow=True,
            )

        # CI Insights Quarantine warning logs
        if not self.mergify_ci.branch_name:
            terminalreporter.write_line(
                "No valid branch name found, unable to setup CI Insights Quarantine",
                yellow=True,
            )

        if self.mergify_ci.quarantined_tests is None:
            terminalreporter.write_line("CI Insights Quarantine could not be setup")
        elif self.mergify_ci.quarantined_tests is not None:
            if self.mergify_ci.quarantined_tests.init_error_msg:
                terminalreporter.write_line(
                    self.mergify_ci.quarantined_tests.init_error_msg, yellow=True
                )
            else:
                terminalreporter.write_line(
                    self.mergify_ci.quarantined_tests.quarantined_tests_report()
                )

        # CI Insights Traces upload logs
        if self.mergify_ci.tracer_provider is None:
            terminalreporter.write_line(
                "Mergify Tracer didn't start for unexpected reason (please contact Mergify support); test results will not be uploaded",
                red=True,
            )
        else:
            try:
                self.mergify_ci.tracer_provider.force_flush()
            except Exception as e:
                terminalreporter.write_line(
                    f"Error while exporting traces: {e}",
                    red=True,
                )
            else:
                terminalreporter.write_line(
                    f"MERGIFY_TEST_RUN_ID={self.mergify_ci.test_run_id}",
                )

            try:
                self.mergify_ci.tracer_provider.shutdown()
            except Exception as e:
                terminalreporter.write_line(
                    f"Error while shutting down the tracer: {e}",
                    red=True,
                )

    @property
    def tracer(self) -> typing.Optional[opentelemetry.trace.Tracer]:
        return self.mergify_ci.tracer

    def pytest_sessionstart(self, session: _pytest.main.Session) -> None:
        if self.tracer:
            traceparent = os.environ.get("MERGIFY_TRACEPARENT")
            if traceparent:
                ctx = TraceContextTextMapPropagator().extract(
                    carrier={"traceparent": traceparent}
                )

            self.session_span = self.tracer.start_span(
                "pytest session start",
                attributes={
                    "test.scope": "session",
                },
                context=ctx if traceparent else None,
            )
        self.has_error = False

    def pytest_collection_finish(self, session: _pytest.main.Session) -> None:
        if self.mergify_ci.flaky_detector:
            self.mergify_ci.flaky_detector.prepare_for_session(session)

    @pytest.hookimpl(hookwrapper=True)
    def pytest_sessionfinish(
        self,
        session: _pytest.main.Session,
    ) -> typing.Generator[None, None, None]:
        if not self.tracer:
            yield
            return

        yield

        self.session_span.set_status(
            opentelemetry.trace.StatusCode.ERROR
            if self.has_error
            else opentelemetry.trace.StatusCode.OK
        )
        self.session_span.end()

    def _get_item_attributes(
        self, item: _pytest.nodes.Item
    ) -> typing.Dict[str, typing.Any]:
        filepath, line_number, testname = item.location
        namespace = testname.replace(item.name, "")
        if namespace.endswith("."):
            namespace = namespace[:-1]

        result = {
            SpanAttributes.CODE_FILEPATH: filepath,
            SpanAttributes.CODE_FUNCTION: item.name,
            SpanAttributes.CODE_LINENO: line_number or 0,
            SpanAttributes.CODE_NAMESPACE: namespace,
            "code.file.path": str(_pytest.pathlib.absolutepath(item.reportinfo()[0])),
            "code.line.number": line_number or 0,
            "test.scope": "case",
        }

        if _should_skip_item(item):
            result["cicd.test.quarantined"] = False
            result["test.case.result.status"] = "skipped"
        else:
            result["cicd.test.quarantined"] = (
                self.mergify_ci.mark_test_as_quarantined_if_needed(item)
            )

        return result

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_protocol(
        self,
        item: _pytest.nodes.Item,
        nextitem: typing.Optional[_pytest.nodes.Item],
    ) -> typing.Optional[bool]:
        # Returning `None` lets pytest continue with its normal test execution
        # flow. Returning `True` means we took care of running the protocol.
        # See:
        # https://docs.pytest.org/en/7.1.x/how-to/writing_hook_functions.html#firstresult
        if not self.tracer:
            return None

        with self.tracer.start_as_current_span(
            name=item.nodeid,
            context=opentelemetry.trace.set_span_in_context(self.session_span),
            attributes=self._get_item_attributes(item),
        ) as current_span:
            distinct_outcomes = set()

            # Execute the initial protocol to register its duration, which lets
            # us calculate the number of reruns.
            for report in _pytest.runner.runtestprotocol(
                item=item, nextitem=nextitem, log=True
            ):
                distinct_outcomes.add(report.outcome)

            if (
                not self.mergify_ci.flaky_detector
                or not self.mergify_ci.flaky_detector.is_rerunning_test(item.nodeid)
            ):
                return True

            self.mergify_ci.flaky_detector.set_test_deadline(
                test=item.nodeid,
                timeout=datetime.timedelta(seconds=timeout_seconds)
                if (timeout_seconds := pytest_timeout._get_item_settings(item).timeout)
                else None,
            )

            if self.mergify_ci.flaky_detector.is_test_too_slow(item.nodeid):
                # We won't be able to detect flakiness if the test is too slow,
                # so we stop here.
                return True

            rerun_count = 0
            while not item.keywords.get("is_last_rerun"):
                item.keywords["is_last_rerun"] = (
                    self.mergify_ci.flaky_detector.is_last_rerun_for_test(item.nodeid)
                )

                # Always execute a last rerun before stopping to properly
                # restore finalizers. Otherwise, it can lead to resource leaks.
                for report in self._reruntestprotocol(item, nextitem):
                    distinct_outcomes.add(report.outcome)

                rerun_count += 1

            if "failed" in distinct_outcomes and "passed" in distinct_outcomes:
                current_span.set_attribute("cicd.test.flaky", True)

            current_span.set_attribute("cicd.test.rerun_count", rerun_count)

        return True

    def _reruntestprotocol(
        self, item: _pytest.nodes.Item, nextitem: typing.Optional[_pytest.nodes.Item]
    ) -> typing.List[_pytest.reports.TestReport]:
        """
        Run the protocol for a rerun of a given test.

        In `new` mode, we log rerun failures to pytest's report to enforce a
        quality gate and prevent merging PRs with new flaky tests. In other
        modes (`unhealthy`), we skip logging to avoid blocking CI, but still
        capture reruns in metrics.
        """

        if not self.mergify_ci.flaky_detector:
            return []

        if self.mergify_ci.flaky_detector.mode == "new":
            return _pytest.runner.runtestprotocol(
                item=item, nextitem=nextitem, log=True
            )

        reports = _pytest.runner.runtestprotocol(
            item=item, nextitem=nextitem, log=False
        )
        for report in reports:
            if report.when != "call":
                item.ihook.pytest_runtest_logreport(report=report)  # Log as usual.
            else:
                # Make rerun visible in the logs by temporarily changing
                # outcome. The goal is to count a potential failure as a rerun
                # instead of a regular failure.
                original_outcome = report.outcome
                report.outcome = "rerun"  # type: ignore[assignment]
                item.ihook.pytest_runtest_logreport(report=report)
                report.outcome = original_outcome

        return reports

    @pytest.hookimpl
    def pytest_report_teststatus(
        self,
        report: _pytest.reports.TestReport,
    ) -> typing.Optional[
        typing.Tuple[
            str, str, typing.Union[str, typing.Tuple[str, typing.Mapping[str, bool]]]
        ]
    ]:
        # https://github.com/pytest-dev/pytest-rerunfailures/blob/master/src/pytest_rerunfailures.py#L622-L625
        if report.outcome == "rerun":  # type: ignore[comparison-overlap]
            return "rerun", "R", ("RERUN", {"yellow": True})  # type: ignore[unreachable]

        return None

    @pytest.hookimpl(tryfirst=True)
    def pytest_runtest_teardown(
        self,
        item: _pytest.nodes.Item,
    ) -> None:
        if (
            not self.mergify_ci.flaky_detector
            or not self.mergify_ci.flaky_detector.is_rerunning_test(item.nodeid)
        ):
            return

        # The goal here is to keep only function-scoped finalizers during
        # reruns and restore higher-scoped finalizers only on the last one.
        if item.keywords.get("is_last_rerun"):
            self.mergify_ci.flaky_detector.restore_item_finalizers(item)
        else:
            self.mergify_ci.flaky_detector.suspend_item_finalizers(item)

    def pytest_exception_interact(
        self,
        node: _pytest.nodes.Node,
        call: _pytest.runner.CallInfo[typing.Any],
        report: _pytest.reports.TestReport,
    ) -> None:
        if self.tracer is None:
            return

        excinfo = call.excinfo

        if excinfo is not None:
            test_span = opentelemetry.trace.get_current_span()

            test_span.set_attributes(
                {
                    SpanAttributes.EXCEPTION_TYPE: str(excinfo.type.__name__),
                    SpanAttributes.EXCEPTION_MESSAGE: str(excinfo.value),
                    SpanAttributes.EXCEPTION_STACKTRACE: str(report.longrepr),
                }
            )
            test_span.set_status(
                opentelemetry.trace.Status(
                    status_code=opentelemetry.trace.StatusCode.ERROR,
                    description=f"{excinfo.type}: {excinfo.value}",
                )
            )

    def pytest_runtest_logreport(self, report: _pytest.reports.TestReport) -> None:
        if self.mergify_ci.flaky_detector:
            self.mergify_ci.flaky_detector.try_fill_metrics_from_report(report)

        if self.tracer is None:
            return

        if report.when != "call":
            return

        if report.outcome is None:
            return  # type: ignore[unreachable]

        if (
            self.mergify_ci.flaky_detector
            and self.mergify_ci.flaky_detector.is_test_rerun(report.nodeid)
        ):
            return

        self._update_current_span_from_report(report)

    def _update_current_span_from_report(
        self, report: _pytest.reports.TestReport
    ) -> None:
        has_error = report.outcome == "failed"
        status_code = (
            opentelemetry.trace.StatusCode.ERROR
            if has_error
            else opentelemetry.trace.StatusCode.OK
        )
        self.has_error |= has_error

        test_span = opentelemetry.trace.get_current_span()
        test_span.set_status(status_code)
        test_span.set_attributes(
            {
                "test.case.result.status": report.outcome,
            }
        )

        if (
            self.mergify_ci.flaky_detector
            and self.mergify_ci.flaky_detector.is_rerunning_test(report.nodeid)
        ):
            test_span.set_attributes({"cicd.test.flaky_detection": True})
            if self.mergify_ci.flaky_detector.mode == "new":
                test_span.set_attributes({"cicd.test.new": True})


def pytest_addoption(parser: _pytest.config.argparsing.Parser) -> None:
    group = parser.getgroup("pytest-mergify", "Mergify support for pytest")
    group.addoption(
        "--mergify-api-url",
        help=(
            "URL of the Mergify API (or set via MERGIFY_API_URL environment variable)"
        ),
    )


def pytest_configure(config: _pytest.config.Config) -> None:
    # NOTE(remyduthu):
    # We are using `isinstance` instead of `get_plugin` because the plugin can
    # be registered with a different name (e.g. `pytester`). It feels safer to
    # check the class name directly.
    for plugin in config.pluginmanager.get_plugins():
        if isinstance(plugin, PytestMergify):
            return

    config.pluginmanager.register(PytestMergify(), name="PytestMergify")


def _should_skip_item(item: _pytest.nodes.Item) -> bool:
    if item.get_closest_marker("skip") is not None:
        return True

    skipif_marker = item.get_closest_marker("skipif")
    if skipif_marker is None:
        return False

    condition = skipif_marker.args[0]
    if not isinstance(condition, str):
        return bool(condition)

    # Mimics how pytest evaluate the conditions
    # https://github.com/pytest-dev/pytest/blob/c5a75f2498c86850c4ce13bcf10d56efc92394a4/src/_pytest/skipping.py#L88
    globals_ = {
        "os": os,
        "sys": sys,
        "platform": platform,
        "config": item.config,
    }
    if hasattr(item, "ihook"):
        for dictionary in reversed(
            item.ihook.pytest_markeval_namespace(config=item.config)
        ):
            if not isinstance(dictionary, Mapping):
                raise ValueError(
                    f"pytest_markeval_namespace() needs to return a dict, got {dictionary!r}"
                )
            globals_.update(dictionary)
    if hasattr(item, "obj"):
        globals_.update(item.obj.__globals__)

    condition_code = compile(
        source=condition,
        filename=f"<{skipif_marker.name} condition>",
        mode="eval",
    )

    # nosemgrep: python.lang.security.audit.eval-detected.eval-detected
    return bool(eval(condition_code, globals_))



from opentelemetry.sdk.resources import Resource, ResourceDetector

from pytest_mergify import utils


class CIResourceDetector(ResourceDetector):
    """Detects OpenTelemetry Resource attributes for GitHub Actions."""

    def detect(self) -> Resource:
        if (provider := utils.get_ci_provider()) is None:
            return Resource({})

        return Resource(
            {
                "cicd.provider.name": provider,
            }
        )



import datetime
import http.server
import os
import re
import socketserver
import threading
import typing
import uuid

import _pytest.pytester
import pytest
import responses
from opentelemetry.sdk import trace
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.sdk.trace.export.in_memory_span_exporter import (
    InMemorySpanExporter,
)

import pytest_mergify
from pytest_mergify import utils

pytest_plugins = ["pytester"]


@pytest.fixture(autouse=True)
def set_api_url(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    # Always override API
    monkeypatch.setenv("MERGIFY_API_URL", "http://localhost:9999")


PytesterWithSpanReturnT = typing.Tuple[
    _pytest.pytester.RunResult, typing.Optional[typing.Dict[str, trace.ReadableSpan]]
]


class PytesterWithSpanT(typing.Protocol):
    def __call__(
        self,
        code: str = ...,
        setenv: typing.Optional[typing.Dict[str, typing.Optional[str]]] = ...,
        quarantined_tests: typing.Optional[typing.List[str]] = None,
    ) -> PytesterWithSpanReturnT: ...


_DEFAULT_PYTESTER_CODE = "def test_pass(): pass"


@pytest.fixture
def pytester_with_spans(
    pytester: _pytest.pytester.Pytester,
    monkeypatch: pytest.MonkeyPatch,
) -> PytesterWithSpanT:
    @responses.activate
    def _run(
        code: str = _DEFAULT_PYTESTER_CODE,
        setenv: typing.Optional[typing.Dict[str, typing.Optional[str]]] = None,
        quarantined_tests: typing.Optional[typing.List[str]] = None,
    ) -> PytesterWithSpanReturnT:
        monkeypatch.delenv("PYTEST_MERGIFY_DEBUG", raising=False)
        monkeypatch.setenv("CI", "true")
        monkeypatch.setenv("_PYTEST_MERGIFY_TEST", "true")

        for k, v in (setenv or {}).items():
            if v is None:
                monkeypatch.delenv(k, raising=False)
            else:
                monkeypatch.setenv(k, v)

        api_url = os.getenv("MERGIFY_API_URL")

        qtest_resp: typing.Dict[str, typing.Any]
        if not quarantined_tests:
            qtest_resp = {"quarantined_tests": []}
        else:
            qtest_resp = {
                "quarantined_tests": [
                    {
                        "id": uuid.uuid4().hex,
                        "test_name": qtest,
                        "reason": "reasonfoobar",
                        "branch": None,
                        "created_at": datetime.datetime.now().isoformat(),
                    }
                    for qtest in quarantined_tests
                ]
            }

        responses.add(
            responses.GET,
            re.compile(rf"{api_url}/v1/ci/.*/repositories/.*/quarantines\?branch=.*"),
            status=200,
            json=qtest_resp,
        )

        full_repository = utils.get_repository_name()
        if full_repository is not None:
            try:
                owner, repo = utils.split_full_repo_name(full_repository)
            except utils.InvalidRepositoryFullNameError:
                pass
            else:
                passthrough = responses.Response(
                    responses.POST,
                    f"{api_url}/v1/ci/{owner}/repositories/{repo}/traces",
                    passthrough=True,
                )
                responses.add(passthrough)

        plugin = pytest_mergify.PytestMergify()
        pytester.makepyfile(code)
        result = pytester.runpytest_inprocess(plugins=[plugin])

        spans_as_dict: typing.Optional[typing.Dict[str, ReadableSpan]]
        if code is _DEFAULT_PYTESTER_CODE:
            result.assert_outcomes(passed=1)
        if isinstance(plugin.mergify_ci.exporter, InMemorySpanExporter):
            spans = plugin.mergify_ci.exporter.get_finished_spans()
            spans_as_dict = {span.name: span for span in spans}
            # Make sure we don't lose spans in the process
            assert len(spans_as_dict) == len(spans)
        else:
            spans_as_dict = None

        return result, spans_as_dict

    return _run


class TestHTTPRequestHandler(http.server.BaseHTTPRequestHandler):
    # Class attribute for the response code, set by the fixture.
    response_code: int = 200

    def do_POST(self) -> None:
        path = self.path[1:].split("/")
        # loozy match, who cares
        if path[0] == "v1" and path[-1] == "traces":
            self.send_response(self.__class__.response_code)
            self.send_header("Content-Type", "text/plain")
            self.end_headers()
            self.wfile.write(b"OK")
        else:
            self.send_response(404)
            self.end_headers()

    def log_message(self, format: str, *args: object) -> None:
        # Override to suppress console logging during tests.
        pass


@pytest.fixture
def http_server(request: pytest.FixtureRequest) -> typing.Generator[str, None, None]:
    # Allow parameterization of the response code via request.param.
    response_code = getattr(request, "param", 200)
    TestHTTPRequestHandler.response_code = response_code

    with socketserver.TCPServer(("", 0), TestHTTPRequestHandler) as httpd:
        host, port = httpd.server_address  # retrieve the actual port
        thread = threading.Thread(target=httpd.serve_forever)
        thread.daemon = True
        thread.start()
        yield f"http://{host!s}:{port}"
        httpd.shutdown()



import datetime
import re
import typing

import _pytest.nodes
import _pytest.pytester
import _pytest.reports
import pytest
import responses

import pytest_mergify
from pytest_mergify import ci_insights
from tests import conftest


def _set_test_environment(
    monkeypatch: pytest.MonkeyPatch,
    mode: typing.Literal["new", "unhealthy"] = "new",
) -> None:
    monkeypatch.setenv("_MERGIFY_TEST_NEW_FLAKY_DETECTION", "true")
    monkeypatch.setenv("_PYTEST_MERGIFY_TEST", "true")
    monkeypatch.setenv("CI", "true")
    monkeypatch.setenv("GITHUB_ACTIONS", "true")
    monkeypatch.setenv("GITHUB_BASE_REF", "main")
    monkeypatch.setenv("GITHUB_REPOSITORY", "Mergifyio/pytest-mergify")
    monkeypatch.setenv("MERGIFY_API_URL", "https://example.com")
    monkeypatch.setenv("MERGIFY_TOKEN", "my_token")

    if mode == "unhealthy":
        # Simulate absence of a PR context: without `GITHUB_BASE_REF` and branch
        # ref variables, `MergifyCIInsights.branch_name` can't be derived,
        # forcing the flaky detector to fall back to `unhealthy` mode. This
        # explicitly exercises the fallback path used when no PR metadata is
        # available.
        monkeypatch.delenv("GITHUB_BASE_REF", raising=False)
        monkeypatch.delenv("GITHUB_HEAD_REF", raising=False)
        monkeypatch.delenv("GITHUB_REF_NAME", raising=False)


def _make_quarantine_mock() -> None:
    responses.add(
        method=responses.GET,
        url="https://example.com/v1/ci/Mergifyio/repositories/pytest-mergify/quarantines",
        json={"quarantined_tests": []},
        status=200,
    )


def _make_flaky_detection_context_mock(
    budget_ratio_for_new_tests: float = 0.1,
    budget_ratio_for_unhealthy_tests: float = 0.05,
    existing_test_names: typing.List[str] = [],
    existing_tests_mean_duration_ms: int = 0,
    unhealthy_test_names: typing.List[str] = [],
    max_test_execution_count: int = 1000,
    max_test_name_length: int = 65536,
    min_budget_duration_ms: int = 4000,
    min_test_execution_count: int = 5,
    status: int = 200,
) -> None:
    responses.add(
        method=responses.GET,
        url="https://example.com/v1/ci/Mergifyio/repositories/pytest-mergify/flaky-detection-context",
        json={
            "budget_ratio_for_new_tests": budget_ratio_for_new_tests,
            "budget_ratio_for_unhealthy_tests": budget_ratio_for_unhealthy_tests,
            "existing_test_names": existing_test_names,
            "existing_tests_mean_duration_ms": existing_tests_mean_duration_ms,
            "unhealthy_test_names": unhealthy_test_names,
            "max_test_execution_count": max_test_execution_count,
            "max_test_name_length": max_test_name_length,
            "min_budget_duration_ms": min_budget_duration_ms,
            "min_test_execution_count": min_test_execution_count,
        },
        status=status,
    )


def _make_test_client() -> ci_insights.MergifyCIInsights:
    return ci_insights.MergifyCIInsights(
        token="my_token",
        repo_name="Mergifyio/pytest-mergify",
        api_url="https://example.com",
    )


@responses.activate
def test_load_flaky_detection(monkeypatch: pytest.MonkeyPatch) -> None:
    _set_test_environment(monkeypatch)

    _make_quarantine_mock()
    _make_flaky_detection_context_mock(existing_test_names=["a::test_a", "b::test_b"])

    client = _make_test_client()
    assert not client.flaky_detector_error_message
    assert client.flaky_detector is not None
    assert client.flaky_detector._context.existing_test_names == [
        "a::test_a",
        "b::test_b",
    ]


@responses.activate
def test_load_flaky_detection_error(monkeypatch: pytest.MonkeyPatch) -> None:
    _set_test_environment(monkeypatch)

    _make_quarantine_mock()
    _make_flaky_detection_context_mock(status=500)

    client = _make_test_client()
    assert client.flaky_detector is None
    assert client.flaky_detector_error_message is not None
    assert "500 Server Error" in client.flaky_detector_error_message


@responses.activate
def test_load_flaky_detection_error_without_existing_tests(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    _set_test_environment(monkeypatch)

    _make_quarantine_mock()
    _make_flaky_detection_context_mock(existing_test_names=[])

    client = _make_test_client()
    assert client.flaky_detector is None
    assert client.flaky_detector_error_message is not None
    assert (
        "No existing tests found for 'Mergifyio/pytest-mergify' repository"
        in client.flaky_detector_error_message
    )


@responses.activate
def test_flaky_detection_for_new_tests(
    monkeypatch: pytest.MonkeyPatch,
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    max_test_name_length = 100

    _set_test_environment(monkeypatch)
    _make_quarantine_mock()
    _make_flaky_detection_context_mock(
        existing_test_names=[
            "test_flaky_detection_for_new_tests.py::test_foo",
            "test_flaky_detection_for_new_tests.py::test_unknown",
        ],
        max_test_name_length=max_test_name_length,
    )

    result, spans = pytester_with_spans(
        code=f"""
        import pytest

        def test_foo():
            assert True

        execution_count = 0

        def test_bar():
            # Simulate a flaky test.
            global execution_count
            execution_count += 1

            if execution_count == 1:
                pytest.fail("I'm flaky!")

        def test_baz():
            assert True

        def test_qux():
            pytest.skip("I'm skipped!")

        def test_quux_{"a" * (max_test_name_length + 10)}():
            assert True

        def test_corge():
            assert True
        """
    )

    result.assert_outcomes(
        failed=1,  # Only the first execution of the flaky test.
        passed=3004,  # The initial execution of the 4 tests and 1000 executions for each new test.
        skipped=1,  # The skipped test is tested only once because skipped tests are excluded from the flaky detection.
    )

    assert re.search(
        r"""🐛 Flaky detection
- Skipped 1 test:
    • 'test_flaky_detection_for_new_tests\.py::test_quux_[a]+' has not been tested multiple times because the name of the test exceeds our limit of \d+ characters
- Used [0-9.]+ % of the budget \([0-9.]+ s/[0-9.]+ s\)
- Active for 3 new tests:
    • 'test_flaky_detection_for_new_tests\.py::test_bar' has been tested \d+ times using approx\. [0-9.]+ % of the budget \([0-9.]+ s/[0-9.]+ s\)
    • 'test_flaky_detection_for_new_tests\.py::test_baz' has been tested \d+ times using approx\. [0-9.]+ % of the budget \([0-9.]+ s/[0-9.]+ s\)
    • 'test_flaky_detection_for_new_tests\.py::test_corge' has been tested \d+ times using approx\. [0-9.]+ % of the budget \([0-9.]+ s/[0-9.]+ s\)""",
        result.stdout.str(),
        re.MULTILINE,
    )

    assert spans is not None
    assert len(spans) == 1 + 6  # 1 for the session and one per test.

    flaky_tests = [
        "test_flaky_detection_for_new_tests.py::test_bar",
    ]
    new_tests = [
        "test_flaky_detection_for_new_tests.py::test_bar",
        "test_flaky_detection_for_new_tests.py::test_baz",
        "test_flaky_detection_for_new_tests.py::test_corge",
    ]
    for span in spans.values():
        assert span is not None
        assert span.attributes is not None

        if span.name in flaky_tests:
            assert span.attributes.get("cicd.test.flaky", False) is True

        if span.name in new_tests:
            assert span.attributes.get("cicd.test.flaky_detection", False) is True
            assert span.attributes.get("cicd.test.new", False) is True
            assert span.attributes.get("cicd.test.rerun_count", 0) == 1000


@responses.activate
def test_flaky_detection_for_unhealthy_tests(
    monkeypatch: pytest.MonkeyPatch,
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    _set_test_environment(monkeypatch, mode="unhealthy")
    _make_quarantine_mock()
    _make_flaky_detection_context_mock(
        unhealthy_test_names=[
            "test_flaky_detection_for_unhealthy_tests.py::test_bar",
            "test_flaky_detection_for_unhealthy_tests.py::test_baz",
            "test_flaky_detection_for_unhealthy_tests.py::test_qux",
            "test_flaky_detection_for_unhealthy_tests.py::test_quux",
            "test_flaky_detection_for_unhealthy_tests.py::test_unknown",
        ],
    )

    result, spans = pytester_with_spans(
        code="""
        import pytest

        def test_foo():
            assert True

        execution_count = 0

        def test_bar():
            # Simulate a flaky test.
            global execution_count
            execution_count += 1

            if execution_count == 2:
                pytest.fail("I'm flaky!")

        def test_baz():
            assert True

        def test_qux():
            pytest.skip("I'm skipped!")

        def test_quux():
            assert True
        """
    )

    # The goal is to make sure the failed rerun of the flaky test does not
    # impact the results to avoid failing the CI of our users.
    assert result.ret == 0

    outcomes = result.parseoutcomes()
    assert len(outcomes) == 3
    assert outcomes["passed"] == 4  # Initial run of each test.
    assert outcomes["skipped"] == 1
    assert outcomes["rerun"] == 3000  # 1000 reruns for each unhealthy test.

    assert re.search(
        r"""🐛 Flaky detection
- Used [0-9.]+ % of the budget \([0-9.]+ s/[0-9.]+ s\)
- Active for 3 unhealthy tests:
    • 'test_flaky_detection_for_unhealthy_tests\.py::test_bar' has been tested \d+ times using approx\. [0-9.]+ % of the budget \([0-9.]+ s/[0-9.]+ s\)
    • 'test_flaky_detection_for_unhealthy_tests\.py::test_baz' has been tested \d+ times using approx\. [0-9.]+ % of the budget \([0-9.]+ s/[0-9.]+ s\)
    • 'test_flaky_detection_for_unhealthy_tests\.py::test_quux' has been tested \d+ times using approx\. [0-9.]+ % of the budget \([0-9.]+ s/[0-9.]+ s\)""",
        result.stdout.str(),
        re.MULTILINE,
    )

    assert spans is not None
    assert len(spans) == 5 + 1  # 1 for the session and one per test.

    flaky_tests = ["test_flaky_detection_for_unhealthy_tests.py::test_bar"]
    unhealthy_tests = [
        "test_flaky_detection_for_unhealthy_tests.py::test_bar",
        "test_flaky_detection_for_unhealthy_tests.py::test_baz",
        "test_flaky_detection_for_unhealthy_tests.py::test_quux",
    ]
    for span in spans.values():
        assert span is not None
        assert span.attributes is not None

        if span.name in flaky_tests:
            assert span.attributes.get("cicd.test.flaky", False) is True

        if span.name in unhealthy_tests:
            assert not span.attributes.get("cicd.test.new")
            assert span.attributes.get("cicd.test.flaky_detection", False) is True
            assert span.attributes.get("cicd.test.rerun_count", 0) == 1000
            # The status should reflect the initial run outcome, not "rerun"
            assert span.attributes.get("test.case.result.status") == "passed"


@responses.activate
def test_flaky_detection_with_fixtures(
    monkeypatch: pytest.MonkeyPatch,
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    max_test_name_length = 100

    _set_test_environment(monkeypatch)
    _make_quarantine_mock()
    _make_flaky_detection_context_mock(
        existing_test_names=[
            "test_flaky_detection_with_fixtures.py::test_first",
            "test_flaky_detection_with_fixtures.py::test_last",
        ],
        max_test_name_length=max_test_name_length,
    )

    suspended_calls, restored_calls = [], []

    from pytest_mergify import flaky_detection

    original_suspend = flaky_detection.FlakyDetector.suspend_item_finalizers
    original_restore = flaky_detection.FlakyDetector.restore_item_finalizers

    def tracked_suspend_item_finalizers(
        self: flaky_detection.FlakyDetector, item: _pytest.nodes.Item
    ) -> None:
        suspended_calls.append(item.nodeid)
        return original_suspend(self, item)

    def tracked_restore_item_finalizers(
        self: flaky_detection.FlakyDetector, item: _pytest.nodes.Item
    ) -> None:
        restored_calls.append(item.nodeid)
        return original_restore(self, item)

    monkeypatch.setattr(
        flaky_detection.FlakyDetector,
        "suspend_item_finalizers",
        tracked_suspend_item_finalizers,
    )
    monkeypatch.setattr(
        flaky_detection.FlakyDetector,
        "restore_item_finalizers",
        tracked_restore_item_finalizers,
    )

    result, spans = pytester_with_spans(
        code="""
        import pytest

        SESSION_ALREADY_SET = False

        SETUP_COUNT = 0
        TEARDOWN_COUNT = 0

        @pytest.fixture(scope="session", autouse=True)
        def _setup_session() -> None:
            global SESSION_ALREADY_SET
            if SESSION_ALREADY_SET:
                raise RuntimeError("This function should not be called twice")
            SESSION_ALREADY_SET = True

        @pytest.fixture(autouse=True)
        def _setup_test():
            global SETUP_COUNT, TEARDOWN_COUNT

            SETUP_COUNT += 1

            yield

            TEARDOWN_COUNT += 1

        def test_first():
            assert True

        # This is a new test.
        def test_second():
            assert True

        def test_last():
            # This test validates that fixtures are properly set up and torn down
            # during test reruns. With 3 tests total (test_first, test_second, test_last)
            # where test_second is new and gets reran 1000 times:
            # - SETUP_COUNT should be 1003 (1 initial run per test + 1000 reruns of test_second)
            # - TEARDOWN_COUNT should be 1002 (all tests complete except test_last which is currently running)
            # This ensures that function-scoped fixtures execute fresh for each rerun,
            # while session-scoped fixtures run only once (validated by SESSION_ALREADY_SET).
            global SETUP_COUNT, TEARDOWN_COUNT
            assert SETUP_COUNT == 1003
            assert TEARDOWN_COUNT == 1002  # Teardown hasn't run yet for test_last.
        """
    )

    result.assert_outcomes(
        passed=1003,  # The initial execution of the 3 tests and 1000 executions for the new test.
    )

    # We should only suspend and restore finalized for the tracked test.
    assert len(suspended_calls) == 1000
    assert all(
        call == "test_flaky_detection_with_fixtures.py::test_second"
        for call in suspended_calls
    )
    assert len(restored_calls) == 1
    assert restored_calls[0] == "test_flaky_detection_with_fixtures.py::test_second"


@responses.activate
def test_flaky_detection_with_only_one_new_test_at_the_end(
    monkeypatch: pytest.MonkeyPatch,
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    _set_test_environment(monkeypatch)
    _make_quarantine_mock()
    _make_flaky_detection_context_mock(
        existing_test_names=[
            "test_flaky_detection_with_only_one_new_test_at_the_end.py::test_foo",
        ]
    )

    result, spans = pytester_with_spans(
        code="""
        import pytest

        SESSION_ALREADY_SET = False

        @pytest.fixture(scope="session", autouse=True)
        def _setup_session() -> None:
            global SESSION_ALREADY_SET
            if SESSION_ALREADY_SET:
                raise RuntimeError("This function should not be called twice")
            SESSION_ALREADY_SET = True

        def test_foo():
            assert True

        def test_bar():
            assert True
        """
    )
    result.assert_outcomes(passed=1002)

    assert spans is not None
    assert len(spans) == 1 + 2  # 1 for the session and one per test.

    span = spans.get(
        "test_flaky_detection_with_only_one_new_test_at_the_end.py::test_bar"
    )
    assert span is not None
    assert span.attributes is not None
    assert span.attributes.get("cicd.test.flaky_detection", False) is True
    assert span.attributes.get("cicd.test.new", False) is True
    assert span.attributes.get("cicd.test.rerun_count", 0) == 1000


@responses.activate
def test_flaky_detection_slow_test_not_reran(
    monkeypatch: pytest.MonkeyPatch,
    pytester: _pytest.pytester.Pytester,
) -> None:
    "Test that a slow test is not reran when it can't reach 5 within the budget."
    _set_test_environment(monkeypatch)
    _make_quarantine_mock()
    _make_flaky_detection_context_mock(
        existing_test_names=[
            "test_flaky_detection_slow_test_not_reran.py::test_existing",
        ],
        min_test_execution_count=5,
    )

    class CustomPlugin:
        def pytest_runtest_makereport(
            self,
            item: _pytest.nodes.Item,
            call: _pytest.reports.TestReport,
        ) -> None:
            if call.when != "call":
                return

            if "test_slow" in item.nodeid:
                call.duration = 10.0  # Simulate a slow test.
            else:
                call.duration = 0.001

    pytester.makepyfile(
        """
        def test_existing():
            assert True

        def test_fast():
            assert True

        def test_slow():
            assert True
        """
    )

    result = pytester.runpytest_inprocess(
        plugins=[CustomPlugin(), pytest_mergify.PytestMergify()]
    )
    result.assert_outcomes(passed=1003)

    # `test_fast` should have been tested successfully.
    assert re.search(
        r"'test_flaky_detection_slow_test_not_reran\.py::test_fast' has been tested \d+ times",
        result.stdout.str(),
    )

    assert (
        "'test_flaky_detection_slow_test_not_reran.py::test_slow' is too slow to be tested at least 5 times within the budget"
        in result.stdout.str()
    )


@responses.activate
def test_flaky_detection_budget_deadline_stops_reruns(
    monkeypatch: pytest.MonkeyPatch,
    pytester: _pytest.pytester.Pytester,
) -> None:
    """
    Test that reruns are stopped when they would exceed the budget deadline.
    """
    _set_test_environment(monkeypatch)
    _make_quarantine_mock()
    _make_flaky_detection_context_mock(
        existing_test_names=[
            "test_flaky_detection_budget_deadline_stops_reruns.py::test_existing",
        ]
    )

    class CustomPlugin:
        deadline_patched: bool = False
        execution_count: int = 0

        def pytest_runtest_call(self, item: _pytest.nodes.Item) -> None:
            plugin = None
            for existing in item.session.config.pluginmanager.get_plugins():
                if isinstance(existing, pytest_mergify.PytestMergify):
                    plugin = existing

            if not plugin or not plugin.mergify_ci.flaky_detector:
                return

            self.execution_count += 1

            # Simulate a slow execution that reaches the deadline.
            if not self.deadline_patched and self.execution_count == 10:
                # Set the deadline in the past to stop after one last rerun.
                plugin.mergify_ci.flaky_detector._test_metrics[
                    "test_flaky_detection_budget_deadline_stops_reruns.py::test_new"
                ].deadline = datetime.datetime.now(
                    datetime.timezone.utc
                ) - datetime.timedelta(hours=1)

                self.deadline_patched = True

    pytester.makepyfile(
        """
        def test_existing():
            assert True

        def test_new():
            assert True
        """
    )

    result = pytester.runpytest_inprocess(
        plugins=[pytest_mergify.PytestMergify(), CustomPlugin()]
    )

    # We should have:
    # - 1 execution of `test_existing`,
    # - 1 initial execution of `test_new`,
    # - Only 9 reruns of `test_new` before the deadline is reached.
    result.assert_outcomes(passed=11)

    assert re.search(
        r"'test_flaky_detection_budget_deadline_stops_reruns\.py::test_new' has been tested 10 times using approx\. [0-9.]+ % of the budget \([0-9.]+ s/[0-9.]+ s\)",
        result.stdout.str(),
    )


@responses.activate
def test_flaky_detector_prepare_for_session_in_new_mode(
    monkeypatch: pytest.MonkeyPatch,
    pytester: _pytest.pytester.Pytester,
) -> None:
    _set_test_environment(monkeypatch, mode="new")
    _make_quarantine_mock()
    _make_flaky_detection_context_mock(
        budget_ratio_for_new_tests=0.5,
        existing_test_names=[
            "test_flaky_detector_prepare_for_session_in_new_mode.py::test_foo",
            "test_flaky_detector_prepare_for_session_in_new_mode.py::test_baz",  # Unknown test, should be filtered.
        ],
        existing_tests_mean_duration_ms=10000,
        max_test_execution_count=10,
    )

    pytester.makepyfile(
        """
        def test_foo():
            assert True

        def test_bar():
            assert True
        """
    )

    plugin = pytest_mergify.PytestMergify()

    result = pytester.runpytest_inprocess(plugins=[plugin])
    result.assert_outcomes(passed=12)  # 2 tests and 10 reruns for the new test.

    assert plugin.mergify_ci.flaky_detector is not None

    # Only the known new test should be in the tests to process.
    assert plugin.mergify_ci.flaky_detector._tests_to_process == [
        "test_flaky_detector_prepare_for_session_in_new_mode.py::test_bar"
    ]
    assert (
        plugin.mergify_ci.flaky_detector._available_budget_duration.total_seconds()
        == datetime.timedelta(seconds=5).total_seconds()
    )


@responses.activate
def test_flaky_detector_prepare_for_session_in_unhealthy_mode(
    monkeypatch: pytest.MonkeyPatch,
    pytester: _pytest.pytester.Pytester,
) -> None:
    _set_test_environment(monkeypatch, mode="unhealthy")
    _make_quarantine_mock()
    _make_flaky_detection_context_mock(
        budget_ratio_for_unhealthy_tests=0.5,
        existing_tests_mean_duration_ms=10000,
        existing_test_names=[
            "test_flaky_detector_prepare_for_session_in_unhealthy_mode.py::test_foo",
            "test_flaky_detector_prepare_for_session_in_unhealthy_mode.py::test_baz",  # Unknown test, should be filtered.
        ],
        unhealthy_test_names=[
            "test_flaky_detector_prepare_for_session_in_unhealthy_mode.py::test_foo",
            "test_flaky_detector_prepare_for_session_in_unhealthy_mode.py::test_baz",  # Unknown test, should be filtered.
        ],
        max_test_execution_count=10,
    )

    pytester.makepyfile(
        """
        def test_foo():
            assert True

        def test_bar():
            assert True
        """
    )

    plugin = pytest_mergify.PytestMergify()

    assert pytester.runpytest_inprocess(plugins=[plugin]).parseoutcomes() == {
        "passed": 2,
        "rerun": 10,
    }

    assert plugin.mergify_ci.flaky_detector is not None

    # Only the known unhealthy test should be in the tests to process.
    assert plugin.mergify_ci.flaky_detector._tests_to_process == [
        "test_flaky_detector_prepare_for_session_in_unhealthy_mode.py::test_foo"
    ]
    assert (
        plugin.mergify_ci.flaky_detector._available_budget_duration.total_seconds()
        == datetime.timedelta(seconds=5).total_seconds()
    )



import _pytest.config
import pytest
from _pytest.pytester import Pytester

import pytest_mergify
from tests import conftest


def test_plugin_is_loaded(pytestconfig: _pytest.config.Config) -> None:
    plugin = pytestconfig.pluginmanager.get_plugin("pytest_mergify")
    assert plugin is pytest_mergify

    plugin = pytestconfig.pluginmanager.get_plugin("PytestMergify")
    assert isinstance(plugin, pytest_mergify.PytestMergify)


def test_no_ci(pytester_with_spans: conftest.PytesterWithSpanT) -> None:
    result, spans = pytester_with_spans(setenv={"CI": "false"})
    assert spans is None
    assert all("Mergify" not in line for line in result.stdout.lines)


@pytest.mark.parametrize("env", ("PYTEST_MERGIFY_ENABLED", "CI"))
def test_enabled(
    env: str,
    pytester_with_spans: conftest.PytesterWithSpanT,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.delenv("CI", raising=False)
    result, spans = pytester_with_spans(setenv={env: "true"})
    assert spans is not None
    assert any("Mergify CI" in line for line in result.stdout.lines)


def test_empty_token(pytester_with_spans: conftest.PytesterWithSpanT) -> None:
    result, spans = pytester_with_spans(
        setenv={
            "MERGIFY_TOKEN": "",
            "_PYTEST_MERGIFY_TEST": None,
        }
    )
    assert spans is None
    assert (
        "No token configured for Mergify; test results will not be uploaded"
        in result.stdout.lines
    )


def test_no_token(pytester_with_spans: conftest.PytesterWithSpanT) -> None:
    result, spans = pytester_with_spans(
        setenv={
            "MERGIFY_TOKEN": None,
            "_PYTEST_MERGIFY_TEST": None,
        }
    )
    assert spans is None
    assert (
        "No token configured for Mergify; test results will not be uploaded"
        in result.stdout.lines
    )


@pytest.mark.parametrize("http_server", [200], indirect=True)
def test_with_token_gha(
    pytester: Pytester,
    monkeypatch: pytest.MonkeyPatch,
    http_server: str,
) -> None:
    monkeypatch.setenv("CI", "1")
    monkeypatch.setenv("GITHUB_ACTIONS", "true")
    monkeypatch.setenv("GITHUB_REPOSITORY", "Mergifyio/pytest-mergify")
    monkeypatch.setenv("MERGIFY_TOKEN", "foobar")
    monkeypatch.setenv("MERGIFY_API_URL", http_server)
    pytester.makepyfile(
        """
        def test_foo():
            assert True
        """
    )
    result = pytester.runpytest_subprocess()
    result.assert_outcomes(passed=1)
    for line in result.stdout.lines:
        if line.startswith("MERGIFY_TEST_RUN_ID="):
            _, test_run_id = line.split("=", 2)
            assert len(test_run_id) == 16
            assert len(bytes.fromhex(test_run_id)) == 8
            break
    else:
        pytest.fail("No trace id found")


def test_repo_name_github_actions(
    pytester: Pytester,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv("CI", "true")
    monkeypatch.setenv("GITHUB_ACTIONS", "true")
    monkeypatch.setenv("GITHUB_REPOSITORY", "Mergifyio/pytest-mergify")
    plugin = pytest_mergify.PytestMergify()
    pytester.makepyfile("")
    pytester.runpytest_inprocess(plugins=[plugin])
    assert plugin.mergify_ci.repo_name == "Mergifyio/pytest-mergify"


def test_with_token_empty_repo(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans(
        setenv={
            "GITHUB_ACTIONS": "true",
            "MERGIFY_TOKEN": "x",
            "_PYTEST_MERGIFY_TEST": "false",
            "GITHUB_REPOSITORY": "",
        }
    )
    assert (
        "Unable to determine repository name; test results will not be uploaded"
        in result.stdout.lines
    )


def test_with_token_no_repo(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans(
        setenv={
            "GITHUB_ACTIONS": "true",
            "MERGIFY_TOKEN": "x",
            "_PYTEST_MERGIFY_TEST": "false",
            "GITHUB_REPOSITORY": None,
        }
    )
    assert (
        "Unable to determine repository name; test results will not be uploaded"
        in result.stdout.lines
    )


def test_errors_logs(
    pytester: _pytest.pytester.Pytester,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    # This will try to upload traces, but we don't have a real exporter so it will log errors.
    monkeypatch.delenv("PYTEST_MERGIFY_DEBUG", raising=False)
    monkeypatch.setenv("MERGIFY_TOKEN", "x")
    monkeypatch.setenv("CI", "1")
    monkeypatch.setenv("GITHUB_ACTIONS", "true")
    monkeypatch.setenv("GITHUB_REPOSITORY", "foo/bar")
    monkeypatch.setenv("GITHUB_REF", "main")
    pytester.makepyfile(
        """
        def test_pass():
            pass
        """
    )
    result = pytester.runpytest_subprocess()
    result.assert_outcomes(passed=1)
    assert any(
        line.startswith(
            "Error while exporting traces: HTTPConnectionPool(host='localhost', port=9999): Max retries exceeded with url"
        )
        for line in result.stdout.lines
    )
    assert not any(
        line.startswith("::notice title=Mergify CI::MERGIFY_TEST_RUN_ID=")
        for line in result.stdout.lines
    )


@pytest.mark.parametrize("http_server", [403], indirect=True)
def test_errors_logs_403(
    pytester: _pytest.pytester.Pytester,
    monkeypatch: pytest.MonkeyPatch,
    http_server: str,
) -> None:
    # This will try to upload traces, but we don't have a real exporter so it will log errors.
    monkeypatch.delenv("PYTEST_MERGIFY_DEBUG", raising=False)
    monkeypatch.setenv("MERGIFY_TOKEN", "x")
    monkeypatch.setenv("CI", "1")
    monkeypatch.setenv("GITHUB_ACTIONS", "true")
    monkeypatch.setenv("GITHUB_REPOSITORY", "foo/bar")
    monkeypatch.setenv("MERGIFY_API_URL", http_server)
    monkeypatch.setenv("GITHUB_BASE_REF", "main")
    pytester.makepyfile(
        """
        def test_pass():
            pass
        """
    )

    result = pytester.runpytest_subprocess()
    result.assert_outcomes(passed=1)
    assert any(
        line.startswith(
            "Error while exporting traces: 403 Client Error: Forbidden for url:"
        )
        for line in result.stdout.lines
    )
    assert not any(
        line.startswith("::notice title=Mergify CI::MERGIFY_TEST_RUN_ID=")
        for line in result.stdout.lines
    )



from opentelemetry.sdk.resources import Resource, ResourceDetector

from pytest_mergify import utils


class MergifyResourceDetector(ResourceDetector):
    """Detects OpenTelemetry Resource attributes for Mergify fields."""

    OPENTELEMETRY_MERGIFY_MAPPING = {
        "mergify.test.job.name": (str, "MERGIFY_TEST_JOB_NAME"),
    }

    def detect(self) -> Resource:
        return Resource(utils.get_attributes(self.OPENTELEMETRY_MERGIFY_MAPPING))



import pytest
from opentelemetry.sdk.resources import Resource, ResourceDetector


class PytestResourceDetector(ResourceDetector):
    """Detects OpenTelemetry Resource attributes for Pytest."""

    def detect(self) -> Resource:
        return Resource(
            {
                "test.framework": "pytest",
                "test.framework.version": pytest.__version__,
            }
        )



import os
import typing

from opentelemetry.sdk.resources import Resource, ResourceDetector
from opentelemetry.semconv._incubating.attributes import cicd_attributes, vcs_attributes

from pytest_mergify import utils
from pytest_mergify.resources import git

GIT_BRANCH_PREFIXES = ("origin/", "refs/heads/")


def _get_branch() -> typing.Optional[str]:
    branch = os.getenv("GIT_BRANCH")
    if branch:
        # NOTE(sileht): it's not 100% bullet proof but since it's very complicated
        # and unlikely to change/add a remote with Jenkins Git/GitHub plugins,
        # we just handle the most common cases.
        for prefix in GIT_BRANCH_PREFIXES:
            if branch.startswith(prefix):
                return branch[len(prefix) :]
        return branch
    return None


class JenkinsResourceDetector(ResourceDetector):
    """Detects OpenTelemetry Resource attributes for Jenkins."""

    OPENTELEMETRY_JENKINS_MAPPING = {
        cicd_attributes.CICD_PIPELINE_NAME: (str, "JOB_NAME"),
        cicd_attributes.CICD_PIPELINE_TASK_NAME: (str, "JOB_NAME"),
        cicd_attributes.CICD_PIPELINE_RUN_ID: (str, "BUILD_ID"),
        "cicd.pipeline.run.url": (str, "BUILD_URL"),
        "cicd.pipeline.runner.name": (str, "NODE_NAME"),
        vcs_attributes.VCS_REF_HEAD_NAME: (str, _get_branch),
        vcs_attributes.VCS_REF_HEAD_REVISION: (str, "GIT_COMMIT"),
        vcs_attributes.VCS_REPOSITORY_URL_FULL: (str, "GIT_URL"),
        "vcs.repository.name": (
            str,
            lambda: utils.get_repository_name_from_env_url("GIT_URL"),
        ),
    }

    def detect(self) -> Resource:
        if utils.get_ci_provider() != "jenkins":
            return Resource({})

        attributes = utils.get_attributes(
            git.GitResourceDetector.OPENTELEMETRY_GIT_MAPPING
        )
        attributes.update(utils.get_attributes(self.OPENTELEMETRY_JENKINS_MAPPING))

        return Resource(attributes)



import datetime
import typing

import _pytest
import _pytest.reports
import freezegun
import pytest

import pytest_mergify
from pytest_mergify import flaky_detection

_NOW = datetime.datetime(
    year=2025,
    month=1,
    day=1,
    hour=0,
    minute=0,
    second=0,
    tzinfo=datetime.timezone.utc,
)


class InitializedFlakyDetector(flaky_detection.FlakyDetector):
    def __init__(self) -> None:
        self.token = ""
        self.url = ""
        self.full_repository_name = ""
        self.mode = "new"
        self._test_metrics = {}

    def __post_init__(self) -> None:
        pass


def _make_flaky_detection_context(
    budget_ratio_for_new_tests: float = 0,
    budget_ratio_for_unhealthy_tests: float = 0,
    existing_test_names: typing.List[str] = [],
    existing_tests_mean_duration_ms: int = 0,
    unhealthy_test_names: typing.List[str] = [],
    max_test_execution_count: int = 0,
    max_test_name_length: int = 0,
    min_budget_duration_ms: int = 0,
    min_test_execution_count: int = 0,
) -> flaky_detection._FlakyDetectionContext:
    return flaky_detection._FlakyDetectionContext(
        budget_ratio_for_new_tests=budget_ratio_for_new_tests,
        budget_ratio_for_unhealthy_tests=budget_ratio_for_unhealthy_tests,
        existing_test_names=existing_test_names,
        existing_tests_mean_duration_ms=existing_tests_mean_duration_ms,
        unhealthy_test_names=unhealthy_test_names,
        max_test_execution_count=max_test_execution_count,
        max_test_name_length=max_test_name_length,
        min_budget_duration_ms=min_budget_duration_ms,
        min_test_execution_count=min_test_execution_count,
    )


def test_flaky_detector_try_fill_metrics_from_report() -> None:
    def make_report(
        nodeid: str, when: typing.Literal["setup", "call", "teardown"], duration: float
    ) -> _pytest.reports.TestReport:
        return _pytest.reports.TestReport(
            duration=duration,
            keywords={},
            location=("", None, ""),
            longrepr=None,
            nodeid=nodeid,
            outcome="passed",
            when=when,
        )

    detector = InitializedFlakyDetector()
    detector._context = _make_flaky_detection_context(max_test_name_length=100)
    detector._tests_to_process = ["foo"]

    plugin = pytest_mergify.PytestMergify()
    plugin.mergify_ci = pytest_mergify.ci_insights.MergifyCIInsights()
    plugin.mergify_ci.flaky_detector = detector

    plugin.pytest_runtest_logreport(make_report(nodeid="foo", when="setup", duration=1))
    plugin.pytest_runtest_logreport(make_report(nodeid="foo", when="call", duration=2))
    plugin.pytest_runtest_logreport(
        make_report(nodeid="foo", when="teardown", duration=3)
    )

    plugin.pytest_runtest_logreport(make_report(nodeid="foo", when="setup", duration=4))
    plugin.pytest_runtest_logreport(make_report(nodeid="foo", when="call", duration=5))
    plugin.pytest_runtest_logreport(
        make_report(nodeid="foo", when="teardown", duration=6)
    )

    metrics = detector._test_metrics.get("foo")
    assert metrics is not None
    assert metrics.initial_duration == datetime.timedelta(seconds=6)
    assert metrics.rerun_count == 2
    assert metrics.total_duration == datetime.timedelta(seconds=21)


def test_flaky_detector_count_remaining_tests() -> None:
    detector = InitializedFlakyDetector()
    detector.mode = "new"
    detector._tests_to_process = ["foo", "bar", "baz"]
    detector._test_metrics = {
        "foo": flaky_detection._TestMetrics(
            deadline=datetime.datetime.now(datetime.timezone.utc)
        ),
        "bar": flaky_detection._TestMetrics(),
        "baz": flaky_detection._TestMetrics(),
    }
    assert detector._count_remaining_tests() == 2


@freezegun.freeze_time(
    time_to_freeze=datetime.datetime.fromisoformat("2025-01-01T00:00:00+00:00")
)
@pytest.mark.parametrize(
    argnames=("metrics", "expected"),
    argvalues=[
        pytest.param(flaky_detection._TestMetrics(), True, id="Deadline not set"),
        pytest.param(
            flaky_detection._TestMetrics(
                deadline=datetime.datetime.fromisoformat("2025-01-02T00:00:00+00:00"),
                initial_call_duration=datetime.timedelta(seconds=1),
            ),
            False,
            id="Not exceeded",
        ),
        pytest.param(
            flaky_detection._TestMetrics(
                deadline=datetime.datetime.fromisoformat("2025-01-01T00:00:00+00:00"),
                initial_call_duration=datetime.timedelta(),
            ),
            True,
            id="Exceeded by deadline",
        ),
        pytest.param(
            flaky_detection._TestMetrics(
                deadline=datetime.datetime.fromisoformat("2025-01-01T00:00:00+00:00"),
                initial_call_duration=datetime.timedelta(minutes=2),
            ),
            True,
            id="Exceeded by initial duration",
        ),
    ],
)
def test_flaky_detector_will_exceed_test_deadline(
    metrics: flaky_detection._TestMetrics,
    expected: bool,
) -> None:
    assert metrics.will_exceed_deadline() == expected


@pytest.mark.parametrize(
    argnames=(
        "available_budget_duration",
        "test_metrics",
        "expected",
    ),
    argvalues=[
        pytest.param(
            datetime.timedelta(seconds=1),
            {
                "baz": flaky_detection._TestMetrics(
                    total_duration=datetime.timedelta(milliseconds=500)
                ),
            },
            # Total test duration: 2 tests × 2000 ms = 4 s
            # Flaky detection budget: 4 s × 0.25 = 1 s
            # Already used: 500 ms (baz's `total_duration`)
            # Remaining budget: 1 s - 500 ms = 500 ms
            datetime.timedelta(milliseconds=500),
            id="Simple",
        ),
        pytest.param(
            datetime.timedelta(milliseconds=400),
            {
                "baz": flaky_detection._TestMetrics(
                    total_duration=datetime.timedelta(milliseconds=500)
                ),
            },
            datetime.timedelta(),
            id="No more budget",
        ),
    ],
)
def test_flaky_detector_get_remaining_budget_duration(
    available_budget_duration: datetime.timedelta,
    test_metrics: typing.Dict[str, flaky_detection._TestMetrics],
    expected: datetime.timedelta,
) -> None:
    detector = InitializedFlakyDetector()
    detector._available_budget_duration = available_budget_duration
    detector._test_metrics = test_metrics
    assert expected == detector._get_remaining_budget_duration()



import opentelemetry.trace
import anys
from opentelemetry.semconv.trace import SpanAttributes

import pytest

from tests import conftest


def test_span(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans()
    assert spans is not None
    assert set(spans.keys()) == {
        "pytest session start",
        "test_span.py::test_pass",
    }


def test_session_without_traceparent(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans()
    assert spans is not None
    s = spans["pytest session start"]
    assert s.attributes == {"test.scope": "session"}
    assert s.status.status_code == opentelemetry.trace.StatusCode.OK
    assert s.parent is None


def test_session_with_traceparent(
    pytester_with_spans: conftest.PytesterWithSpanT,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setenv(
        "MERGIFY_TRACEPARENT", "00-80e1afed08e019fc1110464cfa66635c-7a085853722dc6d2-01"
    )

    result, spans = pytester_with_spans()
    assert spans is not None
    s = spans["pytest session start"]
    assert s.attributes == {"test.scope": "session"}
    assert s.status.status_code == opentelemetry.trace.StatusCode.OK
    assert s.parent is not None
    assert s.context.trace_id == 0x80E1AFED08E019FC1110464CFA66635C
    assert s.parent.trace_id == 0x80E1AFED08E019FC1110464CFA66635C
    assert s.parent.span_id == 0x7A085853722DC6D2


def test_session_fail(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans("def test_fail(): assert False")
    assert spans is not None
    s = spans["pytest session start"]
    assert s.attributes == {"test.scope": "session"}
    assert s.status.status_code == opentelemetry.trace.StatusCode.ERROR


def test_test(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans()
    assert spans is not None
    session_span = spans["pytest session start"]

    assert spans["test_test.py::test_pass"].attributes == {
        "test.scope": "case",
        "code.function": "test_pass",
        "code.lineno": 0,
        "code.filepath": "test_test.py",
        "code.namespace": "",
        "test.case.result.status": "passed",
        "code.file.path": anys.ANY_STR,
        "code.line.number": 0,
        "cicd.test.quarantined": False,
    }
    assert (
        spans["test_test.py::test_pass"].status.status_code
        == opentelemetry.trace.StatusCode.OK
    )
    assert session_span.context is not None
    assert spans["test_test.py::test_pass"].parent is not None
    assert (
        spans["test_test.py::test_pass"].parent.span_id == session_span.context.span_id
    )


def test_test_failure(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans("def test_error(): assert False, 'foobar'")
    assert spans is not None
    session_span = spans["pytest session start"]

    assert spans["test_test_failure.py::test_error"].attributes == {
        "test.case.result.status": "failed",
        "test.scope": "case",
        "code.function": "test_error",
        "code.lineno": 0,
        "code.filepath": "test_test_failure.py",
        "code.namespace": "",
        SpanAttributes.EXCEPTION_TYPE: "AssertionError",
        SpanAttributes.EXCEPTION_MESSAGE: "foobar\nassert False",
        SpanAttributes.EXCEPTION_STACKTRACE: """>   def test_error(): assert False, 'foobar'
E   AssertionError: foobar
E   assert False

test_test_failure.py:1: AssertionError""",
        "code.file.path": anys.ANY_STR,
        "code.line.number": 0,
        "cicd.test.quarantined": False,
    }
    assert (
        spans["test_test_failure.py::test_error"].status.status_code
        == opentelemetry.trace.StatusCode.ERROR
    )
    assert (
        spans["test_test_failure.py::test_error"].status.description
        == "<class 'AssertionError'>: foobar\nassert False"
    )
    assert session_span.context is not None
    assert spans["test_test_failure.py::test_error"].parent is not None
    assert (
        spans["test_test_failure.py::test_error"].parent.span_id
        == session_span.context.span_id
    )


def test_test_skipped(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans("""
import pytest
def test_skipped():
    pytest.skip('not needed')
""")
    assert spans is not None
    session_span = spans["pytest session start"]

    assert spans["test_test_skipped.py::test_skipped"].attributes == {
        "test.case.result.status": "skipped",
        "test.scope": "case",
        "code.function": "test_skipped",
        "code.lineno": 1,
        "code.filepath": "test_test_skipped.py",
        "code.namespace": "",
        "code.file.path": anys.ANY_STR,
        "code.line.number": 1,
        "cicd.test.quarantined": False,
    }
    assert (
        spans["test_test_skipped.py::test_skipped"].status.status_code
        == opentelemetry.trace.StatusCode.OK
    )
    assert session_span.context is not None
    assert spans["test_test_skipped.py::test_skipped"].parent is not None
    assert (
        spans["test_test_skipped.py::test_skipped"].parent.span_id
        == session_span.context.span_id
    )


@pytest.mark.parametrize(
    "mark",
    [
        "skip",
        "skipif(True, reason='not needed')",
        "skipif(1 + 1, reason='with eval')",
        "skipif('1 + 1', reason='as str')",
        "skipif('sys.version_info.major > 1', reason='not needed')",
    ],
)
def test_mark_skipped(
    mark: str,
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans(f"""
import pytest
@pytest.mark.{mark}
def test_skipped():
    assert False
""")
    assert spans is not None
    session_span = spans["pytest session start"]

    assert spans["test_mark_skipped.py::test_skipped"].attributes == {
        "test.case.result.status": "skipped",
        "test.scope": "case",
        "code.function": "test_skipped",
        "code.lineno": 1,
        "code.filepath": "test_mark_skipped.py",
        "code.namespace": "",
        "code.file.path": anys.ANY_STR,
        "code.line.number": 1,
        "cicd.test.quarantined": False,
    }
    assert (
        spans["test_mark_skipped.py::test_skipped"].status.status_code
        == opentelemetry.trace.StatusCode.UNSET
    )
    assert session_span.context is not None
    assert spans["test_mark_skipped.py::test_skipped"].parent is not None
    assert (
        spans["test_mark_skipped.py::test_skipped"].parent.span_id
        == session_span.context.span_id
    )


def test_mark_not_skipped(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans("""
import pytest
@pytest.mark.skipif(False, reason='not skipped')
def test_not_skipped():
    assert True
""")
    assert spans is not None
    session_span = spans["pytest session start"]

    assert spans["test_mark_not_skipped.py::test_not_skipped"].attributes == {
        "test.case.result.status": "passed",
        "test.scope": "case",
        "code.function": "test_not_skipped",
        "code.lineno": 1,
        "code.filepath": "test_mark_not_skipped.py",
        "code.namespace": "",
        "code.file.path": anys.ANY_STR,
        "code.line.number": 1,
        "cicd.test.quarantined": False,
    }
    assert (
        spans["test_mark_not_skipped.py::test_not_skipped"].status.status_code
        == opentelemetry.trace.StatusCode.OK
    )
    assert session_span.context is not None
    assert spans["test_mark_not_skipped.py::test_not_skipped"].parent is not None
    assert (
        spans["test_mark_not_skipped.py::test_not_skipped"].parent.span_id
        == session_span.context.span_id
    )


def test_span_attributes_namespace(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans("""
import pytest

class TestClassBasic:
    def test_namespace(self):
        assert True

def test_namespace():
    assert True


@pytest.mark.parametrize("hello", ["foo", "bar"])
def test_parametrized(hello):
    assert True
""")
    assert spans is not None

    assert "test_span_attributes_namespace.py::test_namespace" in spans
    assert "test_span_attributes_namespace.py::TestClassBasic::test_namespace" in spans
    assert "test_span_attributes_namespace.py::test_parametrized[foo]" in spans
    assert "test_span_attributes_namespace.py::test_parametrized[bar]" in spans


def test_span_resources_test_run_id(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans()
    assert spans is not None
    assert all(
        isinstance(span.resource.attributes["test.run.id"], str)
        and len(span.resource.attributes["test.run.id"]) == 16
        and int(span.resource.attributes["test.run.id"], 16) > 0
        for span in spans.values()
    )



import typing

from opentelemetry.sdk.resources import Resource, ResourceDetector
from opentelemetry.semconv._incubating.attributes import vcs_attributes

from pytest_mergify import utils


def _get_git_commit() -> typing.Optional[str]:
    return utils.git("rev-parse", "HEAD")


def _get_git_branch() -> typing.Optional[str]:
    return utils.git("rev-parse", "--abbrev-ref", "HEAD")


def _get_git_url() -> typing.Optional[str]:
    return utils.git("config", "--get", "remote.origin.url")


def _get_repository_name() -> typing.Optional[str]:
    repository_url = _get_git_url()
    if repository_url:
        return utils.get_repository_name_from_url(repository_url)
    return None


class GitResourceDetector(ResourceDetector):
    """Detects OpenTelemetry Resource attributes for GitHub Actions."""

    OPENTELEMETRY_GIT_MAPPING = {
        vcs_attributes.VCS_REF_HEAD_NAME: (str, _get_git_branch),
        vcs_attributes.VCS_REF_HEAD_REVISION: (str, _get_git_commit),
        vcs_attributes.VCS_REPOSITORY_URL_FULL: (str, _get_git_url),
        "vcs.repository.name": (str, _get_repository_name),
    }

    def detect(self) -> Resource:
        if utils.get_ci_provider() is None:
            return Resource({})

        attributes = utils.get_attributes(self.OPENTELEMETRY_GIT_MAPPING)
        return Resource(attributes)



from tests import conftest
import opentelemetry.trace


def test_spans_quarantine(
    pytester_with_spans: conftest.PytesterWithSpanT,
) -> None:
    result, spans = pytester_with_spans(
        """
import pytest

def test_my_not_flaky_success_test():
    assert True

def test_my_not_flaky_failure_test():
    assert False

def test_my_very_flaky_failure_test():
    assert False

def test_my_very_flaky_success_test():
    assert True
""",
        setenv={
            "MERGIFY_TOKEN": "foobar",
            "GITHUB_ACTIONS": "true",
            "GITHUB_BASE_REF": "main",
            "GITHUB_REPOSITORY": "foo/bar",
        },
        quarantined_tests=[
            "test_spans_quarantine.py::test_my_very_flaky_failure_test",
            "test_spans_quarantine.py::test_my_very_flaky_success_test",
            "test_spans_quarantine.py::some_other_test",
        ],
    )
    assert spans is not None

    assert "test_spans_quarantine.py::test_my_not_flaky_success_test" in spans
    assert (
        spans[
            "test_spans_quarantine.py::test_my_not_flaky_success_test"
        ].status.status_code
        == opentelemetry.trace.StatusCode.OK
    )
    assert (
        spans["test_spans_quarantine.py::test_my_not_flaky_success_test"].attributes
        is not None
    )
    assert not spans[
        "test_spans_quarantine.py::test_my_not_flaky_success_test"
    ].attributes["cicd.test.quarantined"]

    assert "test_spans_quarantine.py::test_my_not_flaky_failure_test" in spans
    assert (
        spans["test_spans_quarantine.py::test_my_not_flaky_failure_test"].attributes
        is not None
    )
    assert (
        spans[
            "test_spans_quarantine.py::test_my_not_flaky_failure_test"
        ].status.status_code
        == opentelemetry.trace.StatusCode.ERROR
    )

    assert not spans[
        "test_spans_quarantine.py::test_my_not_flaky_failure_test"
    ].attributes["cicd.test.quarantined"]

    assert "test_spans_quarantine.py::test_my_very_flaky_failure_test" in spans
    assert "test_spans_quarantine.py::test_my_very_flaky_success_test" in spans
    assert (
        spans["test_spans_quarantine.py::test_my_very_flaky_failure_test"].attributes
        is not None
    )
    assert (
        spans["test_spans_quarantine.py::test_my_very_flaky_success_test"].attributes
        is not None
    )
    assert (
        spans[
            "test_spans_quarantine.py::test_my_very_flaky_failure_test"
        ].status.status_code
        == opentelemetry.trace.StatusCode.OK
    )

    assert spans[
        "test_spans_quarantine.py::test_my_very_flaky_failure_test"
    ].attributes["cicd.test.quarantined"]

    assert (
        spans[
            "test_spans_quarantine.py::test_my_very_flaky_success_test"
        ].status.status_code
        == opentelemetry.trace.StatusCode.OK
    )
    assert spans[
        "test_spans_quarantine.py::test_my_very_flaky_success_test"
    ].attributes["cicd.test.quarantined"]

    assert """🛡️ Quarantine
- Repository: foo/bar
- Branch: main
- Quarantined tests fetched from API: 3

- 🔒 Quarantined:
    · test_spans_quarantine.py::test_my_very_flaky_failure_test
    · test_spans_quarantine.py::test_my_very_flaky_success_test

- Unused quarantined tests:
    · test_spans_quarantine.py::some_other_test
""" in result.stdout.str()
